/*
 * kernel/power/process.c
 *
 * Copyright (C) 1998-2001 Gabor Kuti <seasons@fornax.hu>
 * Copyright (C) 1998,2001,2002 Pavel Machek <pavel@suse.cz>
 * Copyright (C) 2002-2003 Florent Chabaud <fchabaud@free.fr>
 * Copyright (C) 2002-2006 Nigel Cunningham <nigel@suspend2.net>
 *
 * This file is released under the GPLv2.
 *
 * Freeze_and_free contains the routines software suspend uses to freeze other
 * processes during the suspend cycle and to (if necessary) free up memory in
 * accordance with limitations on the image size.
 *
 * Ideally, the image saved to disk would be an atomic copy of the entire 
 * contents of all RAM and related hardware state. One of the first 
 * prerequisites for getting our approximation of this is stopping the activity
 * of other processes. We can't stop all other processes, however, since some 
 * are needed in doing the I/O to save the image. Freeze_and_free.c contains 
 * the routines that control suspension and resuming of these processes.
 * 
 * Under high I/O load, we need to be careful about the order in which we
 * freeze processes. If we freeze processes in the wrong order, we could
 * deadlock others. The freeze_order array this specifies the order in which
 * critical processes are frozen. All others are suspended after these have
 * entered the refrigerator.
 *
 * Another complicating factor is that freeing memory requires the processes
 * to not be frozen, but at the end of freeing memory, they need to be frozen
 * so that we can be sure we actually have eaten enough memory. This is why
 * freezing and freeing are in the one file. The freezer is not called from
 * the main logic, but indirectly, via the code for eating memory. The eat
 * memory logic is iterative, first freezing processes and checking the stats,
 * then (if necessary) unfreezing them and eating more memory until it looks 
 * like the criteria are met (at which point processes are frozen & stats
 * checked again).
 */

#include <linux/suspend.h>
#include <linux/freezer.h>
#include <linux/module.h>
#include <linux/mount.h>
#include <linux/namespace.h>
#include <linux/buffer_head.h>
#include <linux/hardirq.h>
#include <asm/tlbflush.h>

unsigned long freezer_state = 0;

#if 0
//#ifdef CONFIG_PM_DEBUG
#define freezer_message(msg, a...) do { printk(msg, ##a); } while(0)
#else
#define freezer_message(msg, a...) do { } while(0)
#endif

/* Timeouts when freezing */
#define FREEZER_TOTAL_TIMEOUT (5 * HZ)
#define FREEZER_CHECK_TIMEOUT (HZ / 10)

DECLARE_COMPLETION(kernelspace_thaw);
DECLARE_COMPLETION(userspace_thaw);
static atomic_t nr_userspace_frozen;
static atomic_t nr_kernelspace_frozen;

struct frozen_fs
{
	struct list_head fsb_list;
	struct super_block *sb;
};

LIST_HEAD(frozen_fs_list);

void freezer_make_fses_rw(void)
{
	struct frozen_fs *fs, *next_fs;

	list_for_each_entry_safe(fs, next_fs, &frozen_fs_list, fsb_list) {
		thaw_bdev(fs->sb->s_bdev, fs->sb);

		list_del(&fs->fsb_list);
		kfree(fs);
	}
}

/* 
 * Done after userspace is frozen, so there should be no danger of
 * fses being unmounted while we're in here.
 */
int freezer_make_fses_ro(void)
{
	struct frozen_fs *fs;
	struct super_block *sb;

	/* Generate the list */
	list_for_each_entry(sb, &super_blocks, s_list) {
		if (!sb->s_root || !sb->s_bdev ||
		    (sb->s_frozen == SB_FREEZE_TRANS) ||
		    (sb->s_flags & MS_RDONLY))
			continue;

		fs = kmalloc(sizeof(struct frozen_fs), GFP_ATOMIC);
		fs->sb = sb;
		list_add_tail(&fs->fsb_list, &frozen_fs_list);
	};

	/* Do the freezing in reverse order so filesystems dependant
	 * upon others are frozen in the right order. (Eg loopback
	 * on ext3). */
	list_for_each_entry_reverse(fs, &frozen_fs_list, fsb_list)
		freeze_bdev(fs->sb->s_bdev);

	return 0;
}

/*
 * freezeable
 *
 * Description:	Determine whether a process should be frozen yet.
 * Parameters:	struct task_struct *	The process to consider.
 * 		int			Boolean - 0 = userspace else all.
 * Returns:	int			0 if don't freeze yet, otherwise do.
 */
static int freezeable(struct task_struct * p, int all_freezable)
{
	if ((p == current) || 
	    (p->flags & PF_FROZEN) ||
	    (p->flags & PF_NOFREEZE) ||
	    (p->exit_state == EXIT_ZOMBIE) ||
	    (p->exit_state == EXIT_DEAD) ||
	    (p->state == TASK_STOPPED) ||
	    (p->state == TASK_TRACED) ||
	    (!p->mm && !all_freezable))
		return 0;
	return 1;
}

static void __freeze_process(struct completion *completion_handler,
				atomic_t *nr_frozen)
{
	long save;

	freezer_message("%s (%d) frozen.\n",
			current->comm, current->pid);
	save = current->state;
	
	atomic_inc(nr_frozen);
	wait_for_completion(completion_handler);
	atomic_dec(nr_frozen);
	
	current->state = save;
	freezer_message("%s (%d) leaving freezer.\n",
			current->comm, current->pid);
}

/*
 * Refrigerator
 */
void refrigerator(void)
{
	unsigned long flags;

	might_sleep();

	/* Locking to handle race against waking the process in
	 * freeze threads. */
	spin_lock_irqsave(&current->sighand->siglock, flags);
	frozen_process(current);

	recalc_sigpending();
	spin_unlock_irqrestore(&current->sighand->siglock, flags);

	if (test_freezer_state(FREEZER_ON)) {
		if (current->mm)
			__freeze_process(&userspace_thaw, &nr_userspace_frozen);
		else
			__freeze_process(&kernelspace_thaw,
					&nr_kernelspace_frozen);
	}

	spin_lock_irqsave(&current->sighand->siglock, flags);
	recalc_sigpending();
	current->flags &= ~PF_FROZEN;
	spin_unlock_irqrestore(&current->sighand->siglock, flags);

	return;
}

void thaw_processes(int do_all_threads)
{
	if (do_all_threads) {
		clear_freezer_state(FREEZER_ON);
		clear_freezer_state(ABORT_FREEZING);
	}
	
	clear_freezer_state(FREEZING_COMPLETE);

	complete_all(&kernelspace_thaw);
	while (atomic_read(&nr_kernelspace_frozen) > 0)
		yield();

	init_completion(&kernelspace_thaw);
	freezer_make_fses_rw();

	if (do_all_threads) {
		complete_all(&userspace_thaw);
		while (atomic_read(&nr_userspace_frozen) > 0)
			yield();
		init_completion(&userspace_thaw);
	}
}

/*
 * num_freezeable
 *
 * Description:	Determine how many processes of our type are still to be
 * 		frozen. As a side effect, update the progress bar too.
 * Parameters:	int	Which type we are trying to freeze.
 * 		int	Whether we are displaying our progress.
 */
static int num_freezeable(int do_all_threads) {
	
	struct task_struct *g, *p;
	int todo_this_type = 0;

	read_lock(&tasklist_lock);
	do_each_thread(g, p) {
		if (freezeable(p, do_all_threads))
			todo_this_type++;
	} while_each_thread(g, p);
	read_unlock(&tasklist_lock);

	return todo_this_type;
}

/*
 * num_uninterruptible
 *
 * Description:	Determine how many processes of our type are in state
 * 		task uninterruptible.
 * Parameters:	int	Which type we are trying to freeze.
 */
static int num_uninterruptible(int do_all_threads) {
	
	struct task_struct *g, *p;
	int count = 0;

	read_lock(&tasklist_lock);
	do_each_thread(g, p) {
		if (freezeable(p, do_all_threads) &&
			p->state == TASK_UNINTERRUPTIBLE)
			count++;
	} while_each_thread(g, p);
	read_unlock(&tasklist_lock);

	return count;
}

/*
 * Tell threads of the type to enter the freezer.
 */
static void signal_threads(int do_all_threads)
{
	struct task_struct *g, *p;
	unsigned long flags;

	read_lock(&tasklist_lock);
	do_each_thread(g, p) {
		if (!freezeable(p, do_all_threads))
			continue;

		freeze(p);
		spin_lock_irqsave(&p->sighand->siglock, flags);
		signal_wake_up(p, 0);
		spin_unlock_irqrestore(&p->sighand->siglock, flags);
	} while_each_thread(g, p);
	read_unlock(&tasklist_lock);
}

/*
 * Prod processes that haven't entered the refrigerator yet.
 */
static void prod_processes(int do_all_threads)
{
	struct task_struct *g, *p;
	unsigned long flags;

	read_lock(&tasklist_lock);
	do_each_thread(g, p) {
		if (!freezeable(p, do_all_threads))
			continue;
			
		spin_lock_irqsave(&p->sighand->siglock, flags);
		if (!(p->flags & PF_FROZEN)) {
			recalc_sigpending();
			signal_wake_up(p, 0);
		}
		spin_unlock_irqrestore(&p->sighand->siglock, flags);
	} while_each_thread(g, p);
	read_unlock(&tasklist_lock);
}

/*
 * Freezer failure.
 *
 * Check whether we failed to freeze all the processes that
 * should be frozen. If we find a task that failed to freeze,
 * we give useful information on what failed and how.
 */
static int freezer_failure(int do_all_threads)
{
	int result = 0;
	struct task_struct *g, *p;

	read_lock(&tasklist_lock);
	do_each_thread(g, p) {
		if (!freezeable(p, do_all_threads) || 
				p->state == TASK_UNINTERRUPTIBLE) 
			continue;

		if (!result) {
			printk(KERN_ERR	"Stopping tasks failed.\n");
			printk(KERN_ERR "Tasks that refused to be "
			 "refrigerated and haven't since exited:\n");
			set_freezer_state(ABORT_FREEZING);
			result = 1;
		}

		if ((freezing(p))) {
			printk(" - %s (#%d) signalled but "
				"didn't enter refrigerator.\n",
				p->comm, p->pid);
		} else
			printk(" - %s (#%d) signalled "
				"and todo list empty.\n",
				p->comm, p->pid);
	} while_each_thread(g, p);
	read_unlock(&tasklist_lock);

	return result;
}

/*
 * freeze_threads
 *
 * Freeze a set of threads having particular attributes.
 *
 * Types:
 * 2: User threads.
 * 3: Kernel threads.
 */
static int freeze_threads(int do_all_threads)
{
	int result = 0, still_to_do;
	unsigned long start_time = jiffies;

	if (do_all_threads)	
		freezer_make_fses_ro();

	signal_threads(do_all_threads);

	/* Watch them do it, wake them if they ignore us. */
	do {
		prod_processes(do_all_threads);

		set_task_state(current, TASK_INTERRUPTIBLE);
		schedule_timeout(FREEZER_CHECK_TIMEOUT);

		still_to_do = num_freezeable(do_all_threads) - 
			num_uninterruptible(do_all_threads);

	} while(still_to_do && (!test_freezer_state(ABORT_FREEZING)) &&
		!time_after(jiffies, start_time + FREEZER_TOTAL_TIMEOUT));

	/*
	 * Did we time out? See if we failed to freeze processes as well.
	 *
	 */
	if ((time_after(jiffies, start_time + FREEZER_TOTAL_TIMEOUT))
			&& (still_to_do))
		result = freezer_failure(do_all_threads);

	if (do_all_threads)
		set_freezer_state(FREEZING_COMPLETE);

	BUG_ON(in_atomic());
	
	return 0; 
}

/*
 * freeze_processes - Freeze processes prior to saving an image of memory.
 * 
 * Return value: 0 = success, 1 = faulure.
 */
int freeze_processes(void)
{
	enum system_states old_state = system_state;
	int result = 0;

	if (!test_freezer_state(FREEZER_ON)) {
		/* 
		 * No race. While !FREEZER_ON, processes
		 * won't enter __freeze_process
		 */
		init_completion(&userspace_thaw);
		init_completion(&kernelspace_thaw);
		set_freezer_state(FREEZER_ON);
	}

	/* Now freeze processes that were syncing and are still running */
	if (freeze_threads(0) || (test_freezer_state(ABORT_FREEZING))) {
		result = 1;
		goto out;
	}

	/* Freeze kernel threads */
	if (freeze_threads(1) || (test_freezer_state(ABORT_FREEZING)))
		result = 1;

out:
	system_state = old_state;
	return result;
}

EXPORT_SYMBOL(freezer_state);
EXPORT_SYMBOL(refrigerator);
